# pycfhelpers

[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)
[![Python 3.7+](https://img.shields.io/badge/python-3.7+-blue.svg)](https://www.python.org/downloads/)
[![Development Status](https://img.shields.io/badge/status-beta-orange.svg)](https://pypi.org/project/pycfhelpers/)

**Python SDK wrapper module for CellFrame blockchain platform**

A comprehensive Python library providing helper functions, utilities, and high-level abstractions for working with the CellFrame blockchain ecosystem. This package is essential for CellFrame plugin development and blockchain application integration.

## 🚀 Features

### Core Functionality
- **Blockchain Operations**: Transaction handling, datum processing, emission management
- **Network Integration**: Multi-network support with automatic discovery
- **Wallet Management**: Address parsing, transaction signing, wallet operations
- **Global Database**: Key-value storage with cluster support and notifications
- **HTTP Servers**: Both simple and Tornado-based HTTP server implementations
- **Logging System**: Unified logging with CellFrame SDK integration and fallbacks

### Advanced Features
- **Graceful Degradation**: Works with or without CellFrame SDK installed
- **Dual Import Support**: Backward compatibility + modern namespace imports
- **Professional Versioning**: Single-source version management
- **Comprehensive Testing**: Security, compatibility, and integrity tests
- **Type Safety**: Full type hints and modern Python practices

## 📦 Installation

### From Wheel (Recommended)
```bash
pip install pycfhelpers-1.0.3-py3-none-any.whl
```

### From Source
```bash
git clone https://gitlab.demlabs.net/cellframe/python-cellframe-modules/pycfhelpers.git
cd pycfhelpers
pip install -e .
```

### Dependencies
- **Required**: `base58>=2.0.0`
- **Optional**: `web3>=6.0.0`, `tornado>=6.0.0`
- **Development**: `flake8>=4.0.0`, `pytest>=7.0.0`, `pytest-cov>=4.0.0`

## 🔧 Quick Start

### Basic Usage
```python
import pycfhelpers

# Check version and availability
print(f"Version: {pycfhelpers.__version__}")
print(f"CellFrame SDK Available: {pycfhelpers.CELLFRAME_AVAILABLE}")

# Basic logging
from pycfhelpers.logger import log
log.info("Application started")

# JSON utilities
from pycfhelpers.helpers import json_dump, json_load
data = {"key": "value"}
json_str = json_dump(data)
parsed = json_load(json_str)
```

### Backward Compatibility (Legacy Code)
```python
# Old imports still work
import logger
import helpers
import cellframenet

# Use as before
logger.log.info("Legacy import works")
```

### Modern Namespace Imports (Recommended)
```python
# New structured imports
from pycfhelpers.node.gdb import CFGDBGroup, CFGDBCluster
from pycfhelpers.node.net import CFNet, CFNodeAddress
from pycfhelpers.node.crypto import CFGUUID
from pycfhelpers.common.parsers import parse_cf_v1_address
from pycfhelpers.node.http.simple import SimpleHTTPServer
```

## 📚 API Reference

### Core Modules

#### `pycfhelpers.helpers`
Essential utility functions for blockchain operations:
```python
from pycfhelpers.helpers import (
    json_dump, json_load,           # JSON serialization
    find_tx_out, get_tx_items,      # Transaction utilities
    get_tsd_data, get_tx_outs,      # Data extraction
    ListCache, net_by_name          # Caching and network utils
)
```

#### `pycfhelpers.logger`
Unified logging system with automatic fallbacks:
```python
from pycfhelpers.logger import log

log.debug("Debug message")
log.info("Info message")
log.warning("Warning message")
log.error("Error message")
log.critical("Critical message")
```

#### `pycfhelpers.cellframenet`
High-level CellFrame blockchain operations:
```python
from pycfhelpers.cellframenet import (
    CellframeNetwork,               # Network management
    CellframeEmission,              # Emission handling
    CellframeBaseTransactionDatum,  # Transaction processing
    TSD                             # Typed Section Data constants
)

# Create network instance
network = CellframeNetwork("mainnet", ["main", "zerochain"])

# Work with emissions
emission = network.get_emission_by_tsd({"type": "transfer"})
```

### Node Modules

#### `pycfhelpers.node.gdb` - Global Database
```python
from pycfhelpers.node.gdb import CFGDBGroup, CFGDBCluster

# Dictionary-like interface
group = CFGDBGroup("my_group")
group["key"] = b"value"
value = group["key"]

# Cluster management
cluster = CFGDBCluster(
    name="test_cluster",
    guuid=CFGUUID(),
    grp_mask="0x0001",
    ttl_hour=24,
    root_access_for_owner=True,
    defalt_member_role=CFGDBCluster.MemberRole.USER,
    cluster_role=CFGDBCluster.ClusterRole.AUTONOMIC
)
```

#### `pycfhelpers.node.net` - Network Operations
```python
from pycfhelpers.node.net import CFNet, CFNodeAddress

# Network operations
net = CFNet("mainnet")
address = CFNodeAddress("cf_address_string")
```

#### `pycfhelpers.node.http` - HTTP Servers
```python
from pycfhelpers.node.http.simple import SimpleHTTPServer
from pycfhelpers.node.http.tornado import TornadoHTTPServer

# Simple HTTP server
server = SimpleHTTPServer(port=8080)
server.start()

# Tornado HTTP server with advanced features
tornado_server = TornadoHTTPServer(port=8081)
tornado_server.add_handler("/api", MyHandler)
```

#### `pycfhelpers.common.parsers` - Address Parsing
```python
from pycfhelpers.common.parsers import parse_cf_v1_address

# Parse CellFrame v1 address
version, net_id, sign_id, public_hash, summary_hash, control_hash = parse_cf_v1_address(
    "mHKK8Z2YQ5y4JR8F6rD3oYtGGgG5V4T8xN1pC7rS5jK3Q9L"
)
```

## 🏗️ Architecture

### Project Structure
```
pycfhelpers/
├── __init__.py              # Main package with version management
├── pycfhelpers.py           # Backward compatibility module
├── helpers.py               # Core utility functions
├── logger.py                # Logging system
├── cellframenet.py          # CellFrame blockchain operations
├── contract.py              # Smart contract utilities
├── tornado_http.py          # Tornado HTTP server
├── common/                  # Common utilities
│   ├── parsers.py          # Address and data parsers
│   └── types.py            # Type definitions
├── node/                   # Node-specific modules
│   ├── gdb.py             # Global Database interface
│   ├── net.py             # Network operations
│   ├── crypto.py          # Cryptographic functions
│   ├── consensus.py       # Consensus mechanisms
│   ├── datums.py          # Datum processing
│   ├── items.py           # Item management
│   ├── logging.py         # Node logging
│   ├── http/              # HTTP server implementations
│   │   ├── simple.py      # Simple HTTP server
│   │   └── tornado.py     # Tornado HTTP server
│   └── srv/               # Service modules
│       └── vote.py        # Voting mechanisms
└── tests/                 # Comprehensive test suite
    ├── test_import_compatibility.py
    ├── test_package_integrity.py
    ├── test_security.py
    └── test_security_compliance.py
```

### Import Strategies

#### 1. Legacy Compatibility (Existing Code)
```python
# These imports continue to work unchanged
import helpers
import logger
import cellframenet
from logger import log
```

#### 2. Modern Namespace (New Code)
```python
# Recommended for new development
from pycfhelpers.node.gdb import CFGDBGroup
from pycfhelpers.logger import log
from pycfhelpers.helpers import json_dump
```

#### 3. Mixed Approach
```python
# Combine both as needed
import helpers  # Legacy
from pycfhelpers.node.gdb import CFGDBGroup  # Modern
```

## 🔒 Security Features

- **Input Validation**: All user inputs are validated and sanitized
- **Safe Imports**: Graceful handling of missing dependencies
- **Type Safety**: Full type hints prevent common errors
- **Access Control**: Role-based access for GDB clusters
- **Secure Defaults**: Conservative default configurations

## 🧪 Testing

The project includes comprehensive tests covering:

### Test Categories
- **Import Compatibility**: Ensures both legacy and modern imports work
- **Package Integrity**: Validates package structure and metadata
- **Security Compliance**: Checks for security vulnerabilities
- **Functional Testing**: Tests core functionality

### Running Tests
```bash
# Install development dependencies
pip install -e ".[dev]"

# Run all tests
pytest

# Run with coverage
pytest --cov=. --cov-report=html

# Run specific test categories
pytest tests/test_security.py
pytest tests/test_import_compatibility.py
```

### CI/CD Pipeline
- **Automated Testing**: GitLab CI runs tests on every commit
- **Build Verification**: Ensures wheel packages build correctly
- **Deployment**: Automatic deployment to artifact repository

## 🔧 Development

### Version Management
The project uses professional single-source versioning:

```bash
# Version is managed in pyproject.toml only
# All modules automatically import the correct version

# Check current version
python -c "import pycfhelpers; print(pycfhelpers.__version__)"
```

### Code Quality
- **Linting**: Flake8 with custom configuration
- **Type Checking**: Full type hints throughout
- **Documentation**: Comprehensive docstrings
- **Testing**: 90%+ test coverage

### Building from Source
```bash
# Clean build
rm -rf dist/ build/ *.egg-info/

# Build wheel
python -m build

# Install locally
pip install dist/*.whl
```

## 🌐 CellFrame Integration

### Prerequisites
For full functionality, install CellFrame SDK:
```bash
# CellFrame SDK provides:
# - CellFrame.* modules
# - DAP.* modules
# - Native blockchain operations
```

### Graceful Degradation
When CellFrame SDK is not available:
- Core utilities still work
- Logging falls back to Python logging
- Dummy classes maintain API compatibility
- Clear error messages guide installation

### Plugin Development
```python
# Essential for CellFrame plugins
import pycfhelpers
from pycfhelpers.logger import log
from pycfhelpers.node.gdb import CFGDBGroup

def plugin_init():
    log.info("Plugin initialized")
    storage = CFGDBGroup("plugin_data")
    storage["config"] = b"plugin_config_data"
```

## 📋 Requirements

### Python Versions
- **Minimum**: Python 3.7
- **Recommended**: Python 3.9+
- **Tested**: Python 3.7, 3.8, 3.9, 3.10, 3.11

### Operating Systems
- **Linux**: Primary platform (Ubuntu, Debian, CentOS)
- **macOS**: Supported
- **Windows**: Basic support

### Dependencies
```toml
# Required
base58 >= 2.0.0

# Optional (install as needed)
web3 >= 6.0.0      # For Web3 integration
tornado >= 6.0.0   # For advanced HTTP servers

# Development
flake8 >= 4.0.0
pytest >= 7.0.0
pytest-cov >= 4.0.0
```

## 🤝 Contributing

### Development Setup
```bash
# Clone repository
git clone https://gitlab.demlabs.net/cellframe/python-cellframe-modules/pycfhelpers.git
cd pycfhelpers

# Install in development mode
pip install -e ".[dev]"

# Run tests
pytest
```

### Code Standards
- **Style**: Follow PEP 8 with Flake8 enforcement
- **Types**: Use type hints for all public APIs
- **Documentation**: Comprehensive docstrings required
- **Testing**: Maintain 90%+ test coverage

### Submitting Changes
1. Create feature branch from `main`
2. Implement changes with tests
3. Ensure all tests pass
4. Submit merge request

## 📄 License

This project is licensed under the MIT License - see the [LICENSE](LICENSE) file for details.

```
MIT License
Copyright (c) 2024 Demlabs
```

## 🔗 Links

- **Repository**: [GitLab](https://gitlab.demlabs.net/cellframe/python-cellframe-modules/pycfhelpers)
- **Issues**: [Bug Tracker](https://gitlab.demlabs.net/cellframe/python-cellframe-modules/pycfhelpers/-/issues)
- **CellFrame**: [Official Website](https://cellframe.net/)
- **Documentation**: [Project Wiki](https://gitlab.demlabs.net/cellframe/python-cellframe-modules/pycfhelpers)

## 📞 Support

- **Email**: support@demlabs.net
- **Issues**: Use GitLab issue tracker
- **Community**: CellFrame developer community

---

**Made with ❤️ by Demlabs for the CellFrame ecosystem**